#!/usr/bin/env python
import numpy as np
import pandas as pd
import seaborn as sns

from matplotlib.colors import ListedColormap
from typing import Optional, Union, Dict

from ktplotspy.utils.support import diverging_palette


def plot_cpdb_heatmap(
    adata: "AnnData",  # actually not needed
    pvals: pd.DataFrame,
    celltype_key: str,
    degs_analysis: bool = False,
    log1p_transform: bool = False,
    alpha: float = 0.05,
    linewidths: float = 0.5,
    row_cluster: bool = True,
    col_cluster: bool = True,
    low_col: str = "#104e8b",
    mid_col: str = "#ffdab9",
    high_col: str = "#8b0a50",
    cmap: Optional[Union[str, ListedColormap]] = None,
    title: str = "",
    return_tables: bool = False,
    symmetrical: bool = True,
    **kwargs
) -> Union[sns.matrix.ClusterGrid, Dict]:
    """Plot cellphonedb results as total counts of interactions.

    Parameters
    ----------
    adata : AnnData
        `AnnData` object with the `.obs` storing the `celltype_key`.
        The `.obs_names` must match the first column of the input `meta.txt` used for `cellphonedb`.
    pvals : pd.DataFrame
        Dataframe corresponding to `pvalues.txt` or `relevant_interactions.txt` from cellphonedb.
    celltype_key : str
        Column name in `adata.obs` storing the celltype annotations.
        Values in this column should match the second column of the input `meta.txt` used for `cellphonedb`.
    degs_analysis : bool, optional
        Whether `cellphonedb` was run in `deg_analysis` mode.
    log1p_transform : bool, optional
        Whether to log1p transform the output.
    alpha : float, optional
        P value threshold value for significance.
    linewidths : float, optional
        Width of lines between each cell.
    row_cluster : bool, optional
        Whether to cluster rows.
    col_cluster : bool, optional
        Whether to cluster columns.
    low_col : str, optional
        Low colour in gradient.
    mid_col : str, optional
        Middle colour in gradient.
    high_col : str, optional
        High colour in gradient.
    cmap : Optional[Union[ListedColormap, str]], optional
        Built-in matplotlib colormap names or custom `ListedColormap`
    title : str, optional
        Plot title.
    return_tables : bool, optional
        Whether to return the dataframes storing the interaction network.
    symmetrical : bool, optional
        Whether to return the sum of interactions as symmetrical heatmap.
    **kwargs
        Passed to seaborn.clustermap.


    Returns
    -------
    Union[sns.matrix.ClusterGrid, Dict]
        Either heatmap of cellphonedb interactions or dataframe containing the interaction network.
    """
    all_intr = pvals.copy()
    intr_pairs = all_intr.interacting_pair
    all_int = all_intr.iloc[:, 11 : all_intr.shape[1]].T
    all_int.columns = intr_pairs
    all_count = all_int.melt(ignore_index=False).reset_index()
    if degs_analysis:
        all_count["significant"] = all_count.value == 1
    else:
        all_count["significant"] = all_count.value < alpha
    count1x = all_count[["index", "significant"]].groupby("index").agg({"significant": "sum"})
    tmp = pd.DataFrame([x.split("|") for x in count1x.index])
    count_final = pd.concat([tmp, count1x.reset_index(drop=True)], axis=1)
    count_final.columns = ["SOURCE", "TARGET", "COUNT"]
    if any(count_final.COUNT > 0):
        count_mat = count_final.pivot_table(index="SOURCE", columns="TARGET", values="COUNT")
        count_mat.columns.name, count_mat.index.name = None, None
        count_mat[pd.isnull(count_mat)] = 0
        if symmetrical:
            count_matx = np.triu(count_mat) + np.tril(count_mat.T) + np.tril(count_mat) + np.triu(count_mat.T)
            count_matx[np.diag_indices_from(count_matx)] = np.diag(count_mat)
            count_matx = pd.DataFrame(count_matx)
            count_matx.columns = count_mat.columns
            count_matx.index = count_mat.index
            count_mat = count_matx.copy()
    if log1p_transform:
        count_mat = np.log1p(count_mat)
    if cmap is None:
        colmap = diverging_palette(low=low_col, medium=mid_col, high=high_col)
    else:
        colmap = cmap
    if not return_tables:
        g = sns.clustermap(
            count_mat,
            row_cluster=row_cluster,
            col_cluster=col_cluster,
            linewidths=linewidths,
            tree_kws={"linewidths": 0},
            cmap=colmap,
            **kwargs
        )
        if title != "":
            g.fig.suptitle(title)
        return g
    else:
        if symmetrical:
            all_sum = pd.DataFrame(count_mat.apply(sum, axis=0), columns=["total_interactions"])
        else:
            count_mat = count_mat.T  # so that the table output is the same layout as the plot
            row_sums = pd.DataFrame(count_mat.apply(sum, axis=0), columns=["total_interactions_row"])
            col_sums = pd.DataFrame(count_mat.apply(sum, axis=1), columns=["total_interactions_col"])
            all_sum = pd.concat([row_sums, col_sums], axis=1)
        out = {"count_network": count_mat, "interaction_count": all_sum, "interaction_edges": count_final}
        return out



#!/usr/bin/env python
import numpy as np
import pandas as pd
import re
import scipy.cluster.hierarchy as shc

try:
    from collections.abc import Iterable
except ImportError:  # pragma: no cover
    from collections import Iterable
from collections import Counter
from itertools import count, tee
from matplotlib.colors import ListedColormap
from scipy.sparse import csr_matrix
from typing import Dict, List, Optional


from ktplotspy.utils.settings import DEFAULT_SEP


def set_x_stroke(df: pd.DataFrame, isnull: bool, stroke: int):
    """Set stroke value in dataframe.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe for plotting.
    isnull : bool
        Whether to check for null.
    stroke : int
        Stroke size value to set.
    """
    for i in df.index:
        if isnull:
            nullstatus = pd.isnull(df.at[i, "x_stroke"])
        else:
            nullstatus = pd.notnull(df.at[i, "x_stroke"])
        if nullstatus:
            df.at[i, "x_stroke"] = stroke


def hclust(data: pd.DataFrame, axis: int = 0) -> List:
    """Perform hierarchical clustering on rows or columns.

    Parameters
    ----------
    data : pd.DataFrame
        Dataframe to perform clustering on.
    axis : int, optional
        Index = 0 and columns = 1

    Returns
    -------
    List
        Column name order after hierarchical clustering.
    """
    if data.shape[axis] > 1:
        # TODO (KT): perhaps can pass a kwargs so that linkage can be adjusted?
        if axis == 1:  # pragma: no cover
            data = data.T
        labels = list(data.index)
        data_clusters = shc.linkage(data, method="average", metric="euclidean")
        data_dendrogram = shc.dendrogram(Z=data_clusters, no_plot=True, labels=labels)
        data_order = data_dendrogram["ivl"]
    else:  # pragma: no cover
        if axis == 1:
            labels = data.columns
        else:
            labels = data.index
        # just return the column names as is
        data_order = list(labels)
    return data_order


def filter_interaction_and_celltype(data: pd.DataFrame, genes: List, celltype_pairs: List) -> pd.DataFrame:
    """Filter data to interactions and celltypes.

    Parameters
    ----------
    data : pd.DataFrame
        Input table to perform the filtering.
    genes : List
        List of query genes.
    celltype_pairs : List
        Column names of celltype pairs

    Returns
    -------
    pd.DataFrame
        Filtered dataframe.
    """
    filtered_data = data[data.interacting_pair.isin(genes)][celltype_pairs]
    return filtered_data


def ensure_categorical(meta: pd.DataFrame, key: str) -> pd.DataFrame:
    """Enforce categorical columns.

    Parameters
    ----------
    meta : pd.DataFrame
        Metadata table.
    key : str
        Column key to ensure categorical.

    Returns
    -------
    pd.DataFrame
        Table with formatted column.
    """
    if not is_categorical(meta[key]):
        meta[key] = meta[key].astype("category")
    return meta


def prep_celltype_query(
    meta: pd.DataFrame,
    cell_type1: str,
    cell_type2: str,
    pattern: str,
    split_by: Optional[str] = None,
) -> List:
    """Prepare regex query for celltypes.

    Parameters
    ----------
    meta : pd.DataFrame
        Metadata table.
    cell_type1 : str
        Name of celltype 1.
    cell_type2 : str
        Name of celltype 2.
    pattern : str
        Special character string pattern to substitute.
    split_by : Optional[str], optional
        How to split plotting groups.

    Returns
    -------
    List
        List of celltype querys.
    """
    labels = list(meta._labels.cat.categories)
    if split_by is not None:
        groups = list(meta[split_by].cat.categories)
    ct1 = [l for l in labels if re.search(cell_type1, l)]
    ct2 = [l for l in labels if re.search(cell_type2, l)]
    c_type1 = sub_pattern_loop(ct1, pattern)
    c_type2 = sub_pattern_loop(ct2, pattern)
    celltype = []
    for i in range(0, len(c_type1)):
        cq = create_celltype_query(c_type1[i], c_type2, DEFAULT_SEP)
        if split_by is not None:
            for g in groups:
                cqi = keep_interested_groups(g, cq, DEFAULT_SEP)
                if cqi != "":
                    celltype.append(cqi)
        else:
            celltype.append(cq)
    return celltype


def prep_query_group(means: pd.DataFrame, custom_dict: Optional[Dict[str, List[str]]] = None) -> Dict:
    """Return gene family query groups.

    Parameters
    ----------
    means : pd.DataFrame
        Means table.
    custom_dict : Optional[Dict[str, List[str]]], optional
        If provided, will update the query groups with the custom list of genes.

    Returns
    -------
    Dict
        Dictionary of gene families.
    """
    chemokines = [i for i in means.interacting_pair if re.search(r"^CXC|CCL|CCR|CX3|XCL|XCR", i)]
    th1 = [
        i
        for i in means.interacting_pair
        if re.search(
            r"IL2|IL12|IL18|IL27|IFNG|IL10|TNF$|TNF |LTA|LTB|STAT1|CCR5|CXCR3|IL12RB1|IFNGR1|TBX21|STAT4",
            i,
        )
    ]
    th2 = [i for i in means.interacting_pair if re.search(r"IL4|IL5|IL25|IL10|IL13|AREG|STAT6|GATA3|IL4R", i)]
    th17 = [
        i
        for i in means.interacting_pair
        if re.search(
            r"IL21|IL22|IL24|IL26|IL17A|IL17A|IL17F|IL17RA|IL10|RORC|RORA|STAT3|CCR4|CCR6|IL23RA|TGFB",
            i,
        )
    ]
    treg = [i for i in means.interacting_pair if re.search(r"IL35|IL10|FOXP3|IL2RA|TGFB", i)]
    costimulatory = [
        i
        for i in means.interacting_pair
        if re.search(
            r"CD86|CD80|CD48|LILRB2|LILRB4|TNF|CD2|ICAM|SLAM|LT[AB]|NECTIN2|CD40|CD70|CD27|CD28|CD58|TSLP|PVR|CD44|CD55|CD[1-9]",
            i,
        )
    ]
    coinhibitory = [i for i in means.interacting_pair if re.search(r"SIRP|CD47|ICOS|TIGIT|CTLA4|PDCD1|CD274|LAG3|HAVCR|VSIR", i)]

    query_dict = {
        "chemokines": chemokines,
        "th1": th1,
        "th2": th2,
        "th17": th17,
        "treg": treg,
        "costimulatory": costimulatory,
        "coinhibitory": coinhibitory,
    }
    if custom_dict is not None:
        for k, r in custom_dict.items():
            query_dict.update({k: [i for i in means.interacting_pair if re.search(r"|".join(r), i)]})
    return query_dict


def prep_table(data: pd.DataFrame) -> pd.DataFrame:
    """Generic function to format the means and pvalues tables.

    Parameters
    ----------
    data : pd.DataFrame
        Input dataframe. Either pandas DataFrame for means or pvalues.

    Returns
    -------
    pd.DataFrame
        Table ready for further analysis.
    """
    dat = data.copy()
    dat.index = make_unique(dat.interacting_pair)
    dat.columns = [re.sub("\\|", DEFAULT_SEP, col) for col in dat.columns]
    dat.index = [re.sub("_", "-", row) for row in dat.index]
    dat.index = [re.sub("[.]", " ", row) for row in dat.index]

    return dat


def make_unique(seq: pd.Series) -> List:
    """Make unique names.

    Parameters
    ----------
    seq : pd.Series
        Series to convert to unique.

    Returns
    -------
    List
        List of unique names.
    """
    seq = list(seq)
    not_unique = [k for k, v in Counter(seq).items() if v > 1]  # so we have: ['name', 'zip']
    # suffix generator dict - e.g., {'name': <my_gen>, 'zip': <my_gen>}
    suff_gens = dict(zip(not_unique, tee(count(1), len(not_unique))))
    for idx, s in enumerate(seq):
        try:
            suffix = "_" + str(next(suff_gens[s]))
        except KeyError:
            # s was unique
            continue
        else:
            seq[idx] += suffix
    return seq


def sub_pattern(cell_type: str, pattern: str) -> str:
    """Substitute special characters in celltype name.

    Parameters
    ----------
    cell_type : str
        Celltype name to find and replace special pattern.
    pattern : str
        Special pattern to find and replace.

    Returns
    -------
    str
        Formatted celltype name.
    """
    cell_type_tmp = [*cell_type]
    cell_typex = "".join(["\\" + str(x) if re.search(pattern, x) else str(x) for x in cell_type_tmp])
    return cell_typex


def sub_pattern_loop(cell_types: list, pattern: str) -> List:
    """For-loop to substitute special characters in celltype names.

    Parameters
    ----------
    cell_types : list
        List of celltypes to find and replace special pattern.
    pattern : str
        Special pattern to find and replace.

    Returns
    -------
    List
        List of formatted celltype names.
    """
    celltypes = []
    for c in cell_types:
        cx = sub_pattern(cell_type=c, pattern=pattern)
        celltypes.append(cx)
    return celltypes


def is_categorical(series: pd.Series) -> bool:
    """Check if pandas Series is categorical

    Parameters
    ----------
    series : pd.Series
        Series to check.

    Returns
    -------
    bool
        Whether it is categorical or not.
    """
    return series.dtype.name == "category"


def create_celltype_query(ctype1: str, ctypes2: List, sep: str) -> List:
    """Create a regex string term with celltypes.

    Parameters
    ----------
    ctype1 : str
        Name of celltype 1.
    ctypes2 : List
        List of celltype 2 names.
    sep : str
        Character separator to store the split between celltype1 and celltype2s.

    Returns
    -------
    List
        List of regex patterns for celltype1-celltype2s.
    """
    ct = []
    for cx2 in ctypes2:
        ct.append("^" + ctype1 + sep + cx2 + "$")
        ct.append("^" + cx2 + sep + ctype1 + "$")
    ct = "|".join(ct)
    return ct


def keep_interested_groups(grp: str, ct: str, sep: str) -> str:
    """Filter function to only keep interested group in a regex string pattern.

    Parameters
    ----------
    grp : str
        Pattern for interested group.
    ct : str
        Input regex term
    sep : str
        Character separator to store the split between celltypes of the same group.

    Returns
    -------
    str
        Final regex string pattern.
    """
    ctx = ct.split("|")
    ctx = [x for x in ctx if re.search(grp + ".*" + sep + grp, x)]
    return "|".join(ctx)


def hex_to_rgb(hex: str) -> List:
    """Convert hex code to RGB values.

    e.g. "#FFFFFF" -> [255,255,255]

    Parameters
    ----------
    hex : str
        Hex colour code.

    Returns
    -------
    List
        RGB colour code.
    """
    # Pass 16 to the integer function for change of base
    return [int(hex[i : i + 2], 16) for i in range(1, 6, 2)]


def rgb_to_hex(rgb: List) -> str:
    """Convert RGB values to hex code.

    e.g. [255,255,255] -> "#FFFFFF"

    Parameters
    ----------
    rgb : List
        RGB colour code.

    Returns
    -------
    str
        Hex colour code.
    """
    # Components need to be integers for hex to make sense
    if len(rgb) == 4:
        rgb = [int(x) for x in rgb[:3]]  # pragma: no cover
    else:
        rgb = [int(x) for x in rgb]
    return "#" + "".join(["0{0:x}".format(v) if v < 16 else "{0:x}".format(v) for v in rgb])


def colour_dict(gradient: List) -> Dict:
    """Generate dictionary of colours.


    Takes in a list of RGB sub-lists and returns dictionary of colours in RGB and hex form.

    Parameters
    ----------
    gradient : List
        List of RGB colours.

    Returns
    -------
    Dict
        Dictionary of colours.
    """
    return {
        "hex": [rgb_to_hex(RGB) for RGB in gradient],
        "r": [RGB[0] for RGB in gradient],
        "g": [RGB[1] for RGB in gradient],
        "b": [RGB[2] for RGB in gradient],
    }


def linear_gradient(start_hex: str, finish_hex: str = "#FFFFFF", n: int = 10) -> Dict:
    """Return a gradient list of n colours between two hex colours.

    `start_hex` and `finish_hex` should be the full six-digit colour string, including the number sign ("#FFFFFF").

    Parameters
    ----------
    start_hex : str
        Starting hex colour code.
    finish_hex : str, optional
        Finishing hex colour code.
    n : int, optional
        Number of colours between start and finish.

    Returns
    -------
    Dict
        Dictionary of colour gradient.
    """
    # Starting and ending colours in RGB form
    s = hex_to_rgb(start_hex)
    f = hex_to_rgb(finish_hex)
    # Initilize a list of the output colours with the starting colour
    RGB_list = [s]
    # Calculate a colour at each evenly spaced value of t from 1 to n
    for t in range(1, n):
        # Interpolate RGB vector for colour at the current value of t
        curr_vector = [int(s[j] + (float(t) / (n - 1)) * (f[j] - s[j])) for j in range(3)]
        # Add it to our list of output colours
        RGB_list.append(curr_vector)

    return colour_dict(RGB_list)


def diverging_palette(low: str, medium: str, high: str, n: int = 4096) -> ListedColormap:
    """ListerColormap with diverging palette.

    Parameters
    ----------
    low : str
        Colour for low.
    medium : str
        Colour for middle.
    high : str
        Colour for high.
    n : int, optional
        Number of colours between low and high.

    Returns
    -------
    ListedColormap
        Diverging colour palette.
    """
    newcmp = ListedColormap(
        linear_gradient(start_hex=low, finish_hex=medium, n=int(n / 2))["hex"]
        + linear_gradient(start_hex=medium, finish_hex=high, n=int(n / 2))["hex"]
    )

    return newcmp


def flatten(l: List[List]) -> List:
    """
    Flatten a list-in-list-in-list.

    Parameters
    ----------
    l : List[List]
        a list-in-list list

    Yields
    ------
    List
        a flattened list.
    """
    for el in l:
        if isinstance(el, Iterable) and not isinstance(el, (str, bytes)):
            yield from flatten(el)
        else:
            yield el


def celltype_means(adata: "AnnData", layer: Optional[str] = None) -> np.ndarray:
    """Compute mean of gene expression.

    Parameters
    ----------
    adata : AnnData
        input `AnnData` object.
    layer : Optional[str], optional
        if left as None, will use `.X`.

    Returns
    -------
    np.ndarray
        mean expression.
    """
    if layer is None:
        if isinstance(adata.X, csr_matrix):
            return np.mean(adata.X.toarray(), axis=0)
        else:  # assume it's numpy array
            return np.mean(adata.X, axis=0)  # pragma: no cover
    else:
        if isinstance(adata.layers[layer], csr_matrix):
            return np.mean(adata.layers[layer].toarray(), axis=0)
        else:
            return np.mean(adata.layers[layer], axis=0)


def celltype_fraction(adata: "AnnData", layer: Optional[str] = None) -> np.ndarray:
    """Compute non-zeor expression fraction

    Parameters
    ----------
    adata : AnnData
        input `AnnData` object.
    layer : Optional[str], optional
        if left as None, will use `.X`.

    Returns
    -------
    np.ndarray
        non-zero expression fraction
    """
    if layer is None:
        if isinstance(adata.X, csr_matrix):
            return np.mean(adata.X.toarray() > 0, axis=0)
        else:  # assume it's numpy array
            return np.mean(adata.X > 0, axis=0)  # pragma: no cover
    else:
        if isinstance(adata.layers[layer], csr_matrix):
            return np.mean(adata.layers[layer].toarray() > 0, axis=0)
        else:
            return np.mean(adata.layers[layer] > 0, axis=0)


def present(x) -> bool:
    """Utility function to check if x is not null or blank."""
    return pd.notnull(x) and x != ""


def find_complex(interaction_df: pd.DataFrame) -> List[str]:
    """Return complexes.

    Parameters
    ----------
    interaction_df : pd.DataFrame
        processed mean table.

    Returns
    -------
    List[str]
        list of complexes.
    """
    idxa = [i for i, j in interaction_df.gene_a.items() if not present(j)]
    idxb = [i for i, j in interaction_df.gene_b.items() if not present(j)]
    complexa = [re.sub("complex:", "", x) for x in interaction_df.loc[idxa, "partner_a"]]
    complexb = [re.sub("complex:", "", x) for x in interaction_df.loc[idxb, "partner_b"]]
    if len(complexa) > 0:
        if len(complexb) > 0:
            return complexa + complexb
        else:
            return complexa  # pragma: no cover
    elif len(complexb) > 0:
        return complexb  # pragma: no cover
    else:
        return []


def generate_df(
    interactions_subset: pd.DataFrame,
    cell_type_grid: pd.DataFrame,
    cell_type_means: pd.DataFrame,
    cell_type_fractions: pd.DataFrame,
    sep: str = DEFAULT_SEP,
) -> pd.DataFrame:
    """Generate final dataframe for doing plotting.

    Parameters
    ----------
    interactions_subset : pd.DataFrame
        processed mean table.
    cell_type_grid : pd.DataFrame
        basicall an edge list/table.
    cell_type_means : pd.DataFrame
        expression dataframe.
    cell_type_fractions : pd.DataFrame
        fraction dataframe.
    sep : str, optional
        separator used for making barcodes.

    Returns
    -------
    pd.DataFrame
        final dataframe use for plotting.
    """
    ligand = list(interactions_subset.id_a)
    receptor = list(interactions_subset.id_b)
    pp = list(cell_type_grid.source)
    rc = list(cell_type_grid.target)
    producer_expression = pd.DataFrame(columns=list(set(pp)))
    producer_fraction = pd.DataFrame(columns=list(set(pp)))
    receiver_expression = pd.DataFrame(columns=list(set(rc)))
    receiver_fraction = pd.DataFrame(columns=list(set(rc)))
    for i in pp:
        for j in ligand:
            if any([re.search("^" + j + "$", x) for x in cell_type_means.index]):
                producer_expression.loc[j, i] = cell_type_means.loc[j, i]
                producer_fraction.loc[j, i] = cell_type_fractions.loc[j, i]
            else:  # pragma: no cover
                producer_expression.loc[j, i] = 0
                producer_fraction.loc[j, i] = 0
    for i in rc:
        for j in receptor:
            if any([re.search("^" + j + "$", x) for x in cell_type_means.index]):
                receiver_expression.loc[j, i] = cell_type_means.loc[j, i]
                receiver_fraction.loc[j, i] = cell_type_fractions.loc[j, i]
            else:  # pragma: no cover
                receiver_expression.loc[j, i] = 0
                receiver_fraction.loc[j, i] = 0
    out = []
    for _, (px, rx) in cell_type_grid.iterrows():
        for _, (
            ip,
            ga,
            gb,
            pa,
            pb,
            ra,
            rb,
            cp,
            ia,
            ib,
        ) in interactions_subset.iterrows():
            if ra:
                if rb:
                    _out = [
                        ia,
                        ib,
                        ra,
                        rb,
                        ip,
                        cp,
                        px,
                        rx,
                        producer_expression.loc[ia, px],
                        producer_fraction.loc[ia, px],
                        receiver_expression.loc[ib, rx],
                        receiver_fraction.loc[ib, rx],
                    ]
                else:
                    _out = [
                        ia,
                        ib,
                        ra,
                        rb,
                        ip,
                        cp,
                        px,
                        rx,
                        producer_expression.loc[ia, px],
                        producer_fraction.loc[ia, px],
                        receiver_expression.loc[ib, rx],
                        receiver_fraction.loc[ib, rx],
                    ]
            else:
                if rb:
                    _out = [
                        ia,
                        ib,
                        ra,
                        rb,
                        ip,
                        cp,
                        px,
                        rx,
                        producer_expression.loc[ia, px],
                        producer_fraction.loc[ia, px],
                        receiver_expression.loc[ib, rx],
                        receiver_fraction.loc[ib, rx],
                    ]
                else:  # pragma: no cover
                    _out = [
                        ia,
                        ib,
                        ra,
                        rb,
                        ip,
                        cp,
                        px,
                        rx,
                        producer_expression.loc[ia, px],
                        producer_fraction.loc[ia, px],
                        receiver_expression.loc[ib, rx],
                        receiver_fraction.loc[ib, rx],
                    ]
            out.append(
                pd.DataFrame(
                    _out,
                    index=[
                        "ligand",
                        "receptor",
                        "receptor_a",
                        "receptor_b",
                        "pair",
                        "converted_pair",
                        "producer",
                        "receiver",
                        "producer_expression",
                        "producer_fraction",
                        "receiver_expression",
                        "receiver_fraction",
                    ],
                ).T
            )

    _df = pd.concat(out)
    _df["from"] = [p + sep + l for p, l in zip(_df.producer, _df.ligand)]
    _df["to"] = [r + sep + rr for r, rr in zip(_df.receiver, _df.receptor)]
    _df["barcode"] = [pr + "-" + rr + sep + cp for pr, rr, cp in zip(_df.producer, _df.receiver, _df.converted_pair)]
    _df = _df.reset_index(drop=True)
    for i, j in _df.iterrows():
        if (j["receptor_b"]) and not (j["receptor_a"]):
            lg, rc = j["receptor"], j["ligand"]
            con_pair = lg + "-" + rc
            ra, rb = j["receptor_b"], j["receptor_a"]
            px, rx = j["receiver"], j["producer"]
            pre, prf = j["receiver_expression"], j["receiver_fraction"]
            rce, rcf = j["producer_expression"], j["producer_fraction"]
            tos, frs = j["from"], j["to"]
            _df.at[i, "ligand"] = lg
            _df.at[i, "receptor"] = rc
            _df.at[i, "converted_pair"] = con_pair
            _df.at[i, "receptor_a"] = ra
            _df.at[i, "receptor_b"] = rb
            _df.at[i, "producer"] = px
            _df.at[i, "receiver"] = rx
            _df.at[i, "producer_expression"] = pre
            _df.at[i, "producer_fraction"] = prf
            _df.at[i, "receiver_expression"] = rce
            _df.at[i, "receiver_fraction"] = rcf
            _df.at[i, "from"] = frs
            _df.at[i, "to"] = tos
    return _df



#!/usr/bin/env python
import numpy as np
import pandas as pd
import re

from plotnine import (
    aes,
    element_blank,
    element_rect,
    element_text,
    geom_point,
    ggplot,
    ggtitle,
    guide_colourbar,
    guide_legend,
    guides,
    options,
    scale_colour_continuous,
    scale_colour_manual,
    scale_fill_continuous,
    scale_fill_manual,
    scale_size_continuous,
    theme,
    theme_bw,
)
from typing import List, Literal, Optional, Union, Tuple, Dict

from ktplotspy.utils.settings import DEFAULT_SEP, DEFAULT_SPEC_PAT
from ktplotspy.utils.support import (
    ensure_categorical,
    filter_interaction_and_celltype,
    hclust,
    prep_celltype_query,
    prep_query_group,
    prep_table,
    set_x_stroke,
    sub_pattern,
)


def plot_cpdb(
    adata: "AnnData",
    cell_type1: str,
    cell_type2: str,
    means: pd.DataFrame,
    pvals: pd.DataFrame,
    celltype_key: str,
    degs_analysis: bool = False,
    splitby_key: Optional[str] = None,
    alpha: float = 0.05,
    keep_significant_only: bool = True,
    genes: Optional[str] = None,
    gene_family: Optional[Literal["chemokines", "th1", "th2", "th17", "treg", "costimulatory", "coinhibitory"]] = None,
    custom_gene_family: Optional[Dict[str, List[str]]] = None,
    standard_scale: bool = True,
    cluster_rows: bool = True,
    cmap_name: str = "viridis",
    max_size: int = 8,
    max_highlight_size: int = 3,
    default_style: bool = True,
    highlight_col: str = "#d62728",
    highlight_size: Optional[int] = None,
    special_character_regex_pattern: Optional[str] = None,
    exclude_interactions: Optional[Union[List, str]] = None,
    title: str = "",
    return_table: bool = False,
    figsize: Tuple[Union[int, float], Union[int, float]] = (6.4, 4.8),
) -> Union[ggplot, pd.DataFrame]:
    """Plotting cellphonedb results as a dot plot.

    Parameters
    ----------
    adata : AnnData
        `AnnData` object with the `.obs` storing the `celltype_key` with or without `splitby_key`.
        The `.obs_names` must match the first column of the input `meta.txt` used for `cellphonedb`.
    cell_type1 : str
        Name of cell type 1. Accepts regex pattern.
    cell_type2 : str
        Name of cell type 1. Accepts regex pattern.
    means : pd.DataFrame
        Dataframe corresponding to `means.txt` from cellphonedb.
    pvals : pd.DataFrame
        Dataframe corresponding to `pvalues.txt` or `relevant_interactions.txt` from cellphonedb.
    celltype_key : str
        Column name in `adata.obs` storing the celltype annotations.
        Values in this column should match the second column of the input `meta.txt` used for `cellphonedb`.
    degs_analysis : bool, optional
        Whether `cellphonedb` was run in `deg_analysis` mode.
    splitby_key : Optional[str], optional
        If provided, will attempt to split the output plot/table by groups.
        In order for this to work, the second column of the input `meta.txt` used for `cellphonedb` MUST be this format: {splitby}_{celltype}.
    alpha : float, optional
        P value threshold value for significance.
    keep_significant_only : bool, optional
        Whether or not to trim to significant (p<0.05) hits.
    genes : Optional[str], optional
        If provided, will attempt to plot only interactions containing the specified gene(s).
    gene_family : Optional[Literal["chemokines", "th1", "th2", "th17", "treg", "costimulatory", "coinhibitory"]], optional
        If provided, will attempt to plot a predetermined set of chemokines or genes associated with Th1, Th2, Th17, Treg, costimulatory or coinhibitory molecules.
    custom_gene_family : Optional[Dict[str, List[str]]], optional
        If provided, will update the gene_family dictionary with this custom dictionary.
        Both `gene_family` (name of the custom family) and `custom_gene_family` (dictionary holding this new family)
        must be specified for this to work.
    standard_scale : bool, optional
        Whether or not to scale the mean interaction values from 0 to 1 per receptor-ligand variable.
    cluster_rows : bool, optional
        Whether or not to cluster the rows (interactions).
    cmap_name : str, optional
        Matplotlib built-in colormap names.
    max_size : int, optional
        Maximum size of points in plot.
    max_highlight_size : int, optional
        Maximum highlight size of points in plot.
    default_style : bool, optional
        Whether or not to plot in default style or inspired from `squidpy`'s plotting style.
    highlight_col : str, optional
        Colour of highlights marking significant hits.
    highlight_size : Optional[int], optional
        Size of highlights marking significant hits.
    special_character_regex_pattern : Optional[str], optional
        Regex string pattern to perform substitution.
        This option should not realy be used unless there is really REALLY special characters that you really REALLY want to keep.
        Rather than using this option, the easiest way is to not your celltypes with weird characters.
        Just use alpha numeric characters and underscores if necessary.
    exclude_interactions : Optional[Union[List, str]], optional
        If provided, the interactions will be removed from the output.
    title : str, optional
        Plot title.
    return_table : bool, optional
        Whether or not to return the results as a dataframe.
    figsize : Tuple[Union[int, float], Union[int, float]], optional
        Figure size.

    Returns
    -------
    Union[ggplot, pd.DataFrame]
        Either a plotnine `ggplot` plot or a pandas `Dataframe` holding the results.

    Raises
    ------
    KeyError
        If genes and gene_family are both provided, or wrong key for gene family provided, the error will occur.
    """
    if special_character_regex_pattern is None:
        special_character_regex_pattern = DEFAULT_SPEC_PAT
    swapr = True if (cell_type1 == ".") or (cell_type2 == ".") else False
    # prepare data
    metadata = adata.obs.copy()
    means_mat = prep_table(data=means)
    pvals_mat = prep_table(data=pvals)
    if degs_analysis:
        pvals_mat.iloc[:, 12 : pvals_mat.shape[1]] = 1 - pvals_mat.iloc[:, 12 : pvals_mat.shape[1]]
    # ensure celltypes are ok
    cell_type1 = sub_pattern(cell_type=cell_type1, pattern=special_character_regex_pattern)
    cell_type2 = sub_pattern(cell_type=cell_type2, pattern=special_character_regex_pattern)
    # check for query
    if genes is None:
        if gene_family is not None:
            query_group = prep_query_group(means_mat, custom_gene_family)
            if isinstance(gene_family, list):
                query = []
                for gf in gene_family:
                    if gf.lower() in query_group:
                        for gfg in query_group[gf.lower()]:
                            query.append(gfg)
                    else:
                        raise KeyError("gene_family needs to be one of the following: {}".format(query_group.keys()))
                query = list(set(query))
            else:
                if gene_family.lower() in query_group:
                    query = query_group[gene_family.lower()]
                else:
                    raise KeyError("gene_family needs to be one of the following: {}".format(query_group.keys()))
        else:
            query = [i for i in means_mat.interacting_pair if re.search("", i)]
    elif genes is not None:
        if gene_family is not None:
            raise KeyError("Please specify either genes or gene_family, not both.")
        else:
            query = [i for i in means_mat.interacting_pair if re.search("|".join(genes), i)]
    metadata = ensure_categorical(meta=metadata, key=celltype_key)
    # prepare regex query for celltypes
    if splitby_key is not None:
        metadata = ensure_categorical(meta=metadata, key=splitby_key)
        groups = list(metadata[splitby_key].cat.categories)
        metadata["_labels"] = [s + "_" + c for s, c in zip(metadata[splitby_key], metadata[celltype_key])]
        metadata["_labels"] = metadata["_labels"].astype("category")
        cat_orders = []
        for s in metadata[splitby_key].cat.categories:
            for c in metadata[celltype_key].cat.categories:
                cat_orders.append(s + "_" + c)
        cat_orders = [x for x in cat_orders if x in list(metadata._labels)]
        metadata["_labels"] = metadata["_labels"].cat.reorder_categories(cat_orders)
        celltype = prep_celltype_query(
            meta=metadata,
            cell_type1=cell_type1,
            cell_type2=cell_type2,
            pattern=special_character_regex_pattern,
            split_by=splitby_key,
        )
    else:
        metadata["_labels"] = metadata[celltype_key]
        celltype = prep_celltype_query(
            meta=metadata,
            cell_type1=cell_type1,
            cell_type2=cell_type2,
            pattern=special_character_regex_pattern,
            split_by=splitby_key,
        )
    cell_type = "|".join(celltype)
    # keep cell types
    if swapr:
        ct_columns = [ct for ct in means_mat.columns if re.search(ct, cell_type)]
    else:
        ct_columns = [ct for ct in means_mat.columns if re.search(cell_type, ct)]
    # filter
    means_matx = filter_interaction_and_celltype(data=means_mat, genes=query, celltype_pairs=ct_columns)
    pvals_matx = filter_interaction_and_celltype(data=pvals_mat, genes=query, celltype_pairs=ct_columns)
    # reorder the columns
    col_order = []
    if splitby_key is not None:
        for g in groups:
            for c in means_matx.columns:
                if re.search(g, c):
                    col_order.append(c)
    else:
        col_order = means_matx.columns
    means_matx = means_matx[col_order]
    pvals_matx = pvals_matx[col_order]
    # whether or not to fillter to only significant hits
    if keep_significant_only:
        keep_rows = pvals_matx.apply(lambda r: any(r < alpha), axis=1)
        keep_rows = [r for r, k in keep_rows.items() if k]
        if len(keep_rows) > 0:
            pvals_matx = pvals_matx.loc[keep_rows]
            means_matx = means_matx.loc[keep_rows]
    # reun hierarchical clustering on the rows based on interaction value.
    if cluster_rows:
        if means_matx.shape[0] > 2:
            h_order = hclust(means_matx, axis=0)
            means_matx = means_matx.loc[h_order]
            pvals_matx = pvals_matx.loc[h_order]
    if standard_scale:
        means_matx = means_matx.apply(lambda r: (r - np.min(r)) / (np.max(r) - np.min(r)), axis=1)
    means_matx.fillna(0, inplace=True)
    # prepare final table
    colm = "scaled_means" if standard_scale else "means"
    df = means_matx.melt(ignore_index=False).reset_index()
    df.columns = ["interaction_group", "celltype_group", colm]
    df_pvals = pvals_matx.melt(ignore_index=False).reset_index()
    df_pvals.columns = ["interaction_group", "celltype_group", "pvals"]
    df.celltype_group = [re.sub(DEFAULT_SEP, "-", c) for c in df.celltype_group]
    df["pvals"] = df_pvals["pvals"]
    # set factors
    df.celltype_group = df.celltype_group.astype("category")
    # prepare for non-default style plotting
    for i in df.index:
        if df.at[i, colm] == 0:
            df.at[i, colm] = np.nan
    df["x_means"] = df[colm]
    df["y_means"] = df[colm]
    for i in df.index:
        if df.at[i, "pvals"] < alpha:
            df.at[i, "x_means"] = np.nan
            if df.at[i, "pvals"] == 0:
                df.at[i, "pvals"] = 0.001
        if df.at[i, "pvals"] >= alpha:
            if keep_significant_only:
                df.at[i, "y_means"] = np.nan
    df["x_stroke"] = df["x_means"]
    set_x_stroke(df=df, isnull=False, stroke=0)
    set_x_stroke(df=df, isnull=True, stroke=highlight_size)
    if exclude_interactions is not None:
        if not isinstance(exclude_interactions, list):
            exclude_interactions = [exclude_interactions]
        df = df[~df.interaction_group.isin(exclude_interactions)]
    df["neglog10p"] = abs(-1 * np.log10(df.pvals))
    df["neglog10p"] = [0 if x >= 0.05 else j for x, j in zip(df["pvals"], df["neglog10p"])]
    df["significant"] = ["yes" if x < alpha else np.nan for x in df.pvals]
    if all(pd.isnull(df["significant"])):
        df["significant"] = "no"
        highlight_col = "#FFFFFF"
    if return_table:
        return df
    else:
        # set global figure size
        options.figure_size = figsize
        if highlight_size is not None:
            max_highlight_size = highlight_size
            stroke = df.x_stroke
        else:
            stroke = df.neglog10p
        # plotting
        if default_style:
            g = ggplot(df, aes(x="celltype_group", y="interaction_group", colour="significant", fill=colm, size=colm, stroke=stroke))
        else:
            if all(df["significant"] == "no"):
                g = ggplot(df, aes(x="celltype_group", y="interaction_group", colour="significant", fill=colm, size=colm, stroke=stroke))
                default_style = True
            else:
                highlight_col = "#FFFFFF"  # enforce this
                g = ggplot(df, aes(x="celltype_group", y="interaction_group", colour=colm, fill="significant", size=colm, stroke=stroke))
        g = (
            g
            + geom_point(na_rm=True)
            + theme_bw()
            + theme(
                axis_text_x=element_text(angle=90, hjust=0, colour="#000000"),
                axis_text_y=element_text(colour="#000000"),
                axis_ticks=element_blank(),
                axis_title_x=element_blank(),
                axis_title_y=element_blank(),
                legend_key=element_rect(alpha=0, width=0, height=0),
                legend_direction="vertical",
                legend_box="horizontal",
            )
            + scale_size_continuous(range=(0, max_size), aesthetics=["size"])
            + scale_size_continuous(range=(0, max_highlight_size), aesthetics=["stroke"])
        )
        if default_style:
            g = (
                g
                + scale_colour_manual(values=highlight_col, na_translate=False)
                + guides(
                    fill=guide_colourbar(barwidth=4, label=True, ticks=True, draw_ulim=True, draw_llim=True, order=1),
                    size=guide_legend(
                        reverse=True,
                        order=2,
                    ),
                    stroke=guide_legend(
                        reverse=True,
                        order=3,
                    ),
                )
                + scale_fill_continuous(cmap_name=cmap_name)
            )
        else:
            g = (
                g
                + scale_fill_manual(values=highlight_col, na_translate=False)
                + guides(
                    colour=guide_colourbar(barwidth=4, label=True, ticks=True, draw_ulim=True, draw_llim=True, order=1),
                    size=guide_legend(
                        reverse=True,
                        order=2,
                    ),
                    stroke=guide_legend(
                        reverse=True,
                        order=3,
                    ),
                )
            )
            df2 = df.copy()
            for i in df2.index:
                if df2.at[i, "pvals"] < alpha:
                    df2.at[i, colm] = np.nan
            g = (
                g
                + geom_point(aes(x="celltype_group", y="interaction_group", colour=colm, size=colm), df2, inherit_aes=False, na_rm=True)
                + scale_colour_continuous(cmap_name=cmap_name)
            )
        if highlight_size is not None:
            g = g + guides(stroke=None)
        if title != "":
            g = g + ggtitle(title)
        elif gene_family is not None:
            if isinstance(gene_family, list):
                gene_family = ", ".join(gene_family)
            g = g + ggtitle(gene_family)
        return g



#!/usr/bin/env python
import re
import matplotlib.pyplot as plt

import numpy as np
import pandas as pd

from collections import defaultdict
from itertools import combinations
from matplotlib.lines import Line2D
from matplotlib.colors import LinearSegmentedColormap
from pycircos import Garc, Gcircle
from typing import Optional, Tuple, Dict, Union

from ktplotspy.utils.settings import DEFAULT_SEP
from ktplotspy.utils.support import celltype_fraction, celltype_means, find_complex, flatten, generate_df, present
from ktplotspy.plot import plot_cpdb


def plot_cpdb_chord(
    adata: "AnnData",
    means: pd.DataFrame,
    pvals: pd.DataFrame,
    deconvoluted: pd.DataFrame,
    celltype_key: str,
    face_col_dict: Optional[Dict[str, str]] = None,
    edge_col_dict: Optional[Dict[str, str]] = None,
    edge_cmap: LinearSegmentedColormap = plt.cm.nipy_spectral,
    remove_self: bool = True,
    gap: Union[int, float] = 2,
    scale_lw: Union[int, float] = 10,
    size: Union[int, float] = 50,
    interspace: Union[int, float] = 2,
    raxis_range: Tuple[int, int] = (950, 1000),
    labelposition: Union[int, float] = 80,
    label_visible: bool = True,
    figsize: Tuple[Union[int, float], Union[int, float]] = (8, 8),
    legend_params: Dict = {"loc": "center left", "bbox_to_anchor": (1, 1), "frameon": False},
    layer: Optional[str] = None,
    **kwargs
) -> Gcircle:
    """Plotting cellphonedb results as a chord diagram.

    Parameters
    ----------
    adata : AnnData
        `AnnData` object with the `.obs` storing the `celltype_key` with or without `splitby_key`.
        The `.obs_names` must match the first column of the input `meta.txt` used for `cellphonedb`.
    means : pd.DataFrame
        Dataframe corresponding to `means.txt` from cellphonedb.
    pvals : pd.DataFrame
        Dataframe corresponding to `pvalues.txt` or `relevant_interactions.txt` from cellphonedb.
    deconvoluted : pd.DataFrame
        Dataframe corresponding to `deconvoluted.txt` from cellphonedb.
    celltype_key : str
        Column name in `adata.obs` storing the celltype annotations.
        Values in this column should match the second column of the input `meta.txt` used for `cellphonedb`.
    face_col_dict : Optional[Dict[str, str]], optional
        dictionary of celltype : face colours.
        If not provided, will try and use `.uns` from `adata` if correct slot is present.
    edge_col_dict : Optional[Dict[str, str]], optional
        Dictionary of interactions : edge colours. Otherwise, will use edge_cmap option.
    edge_cmap : LinearSegmentedColormap, optional
        a `LinearSegmentedColormap` to generate edge colors.
    remove_self : bool, optional
        whether to remove self edges.
    gap : Union[int, float], optional
        relative size of gaps between edges on arc.
    scale_lw : Union[int, float], optional
        numeric value to scale width of lines.
    size : Union[int, float], optional
        Width of the arc section. If record is provided, the value is
        instead set by the sequence length of the record. In reality
        the actual arc section width in the resultant circle is determined
        by the ratio of size to the combined sum of the size and interspace
        values of the Garc class objects in the Gcircle class object.
    interspace : Union[int, float], optional
        Distance angle (deg) to the adjacent arc section in clockwise
        sequence. The actual interspace size in the circle is determined by
        the actual arc section width in the resultant circle is determined
        by the ratio of size to the combined sum of the size and interspace
        values of the Garc class objects in the Gcircle class object.
    raxis_range : Tuple[int, int], optional
        Radial axis range where line plot is drawn.
    labelposition : Union[int, float], optional
        Relative label height from the center of the arc section.
    label_visible : bool, optional
        Font size of the label. The default is 10.
    figsize : Tuple[Union[int, float], Union[int, float]], optional
        size of figure.
    legend_params : Dict, optional
        additional arguments for `plt.legend`.
    layer : Optional[str], optional
        slot in `AnnData.layers` to access. If `None`, uses `.X`.
    **kwargs
        passed to `plot_cpdb`.

    Returns
    -------
    Gcircle
        a `Gcircle` object from `pycircos`.
    """
    # assert splitby = False
    splitby_key, return_table = None, True
    # run plot_cpdb
    lr_interactions = plot_cpdb(
        adata=adata,
        means=means,
        pvals=pvals,
        celltype_key=celltype_key,
        return_table=return_table,
        splitby_key=splitby_key,
        **kwargs,
    )
    # do some name wrangling
    subset_clusters = list(set(flatten([x.split("-") for x in lr_interactions.celltype_group])))
    adata_subset = adata[adata.obs[celltype_key].isin(subset_clusters)].copy()
    interactions = means[["interacting_pair", "gene_a", "gene_b", "partner_a", "partner_b", "receptor_a", "receptor_b"]].copy()
    interactions["converted"] = [re.sub("-", " ", x) for x in interactions.interacting_pair]
    interactions["converted"] = [re.sub("_", "-", x) for x in interactions.interacting_pair]
    lr_interactions["barcode"] = [a + DEFAULT_SEP + b for a, b in zip(lr_interactions.celltype_group, lr_interactions.interaction_group)]
    interactions_subset = interactions[interactions["converted"].isin(list(lr_interactions.interaction_group))].copy()
    # handle complexes gently
    tm0 = {kx: rx.split("_") for kx, rx in interactions_subset.interacting_pair.items()}
    if any([len(x) > 2 for x in tm0.values()]):
        complex_id, simple_id = [], []
        for i, j in tm0.items():
            if len(j) > 2:
                complex_id.append(i)
            elif len(j) == 2:
                simple_id.append(i)
        _interactions_subset = interactions_subset.loc[complex_id].copy()
        _interactions_subset_simp = interactions_subset.loc[simple_id].copy()
        complex_idx1 = [i for i, j in _interactions_subset.partner_b.items() if re.search("complex:", j)]
        complex_idx2 = [i for i, j in _interactions_subset.partner_a.items() if re.search("complex:", j)]
        # complex_idx
        simple_1 = list(_interactions_subset.loc[complex_idx1, "interacting_pair"])
        simple_2 = list(_interactions_subset.loc[complex_idx2, "interacting_pair"])
        partner_1 = [re.sub("complex:", "", b) for b in _interactions_subset.loc[complex_idx1, "partner_b"]]
        partner_2 = [re.sub("complex:", "", a) for a in _interactions_subset.loc[complex_idx2, "partner_a"]]
        for i, _ in enumerate(simple_1):
            simple_1[i] = re.sub(partner_1[i] + "_|_" + partner_1[i], "", simple_1[i])
        for i, _ in enumerate(simple_2):
            simple_2[i] = re.sub(partner_2[i] + "_|_" + partner_2[i], "", simple_2[i])
        tmpdf = pd.concat([pd.DataFrame(zip(simple_1, partner_1)), pd.DataFrame(zip(partner_2, simple_2))])
        tmpdf.index = complex_id
        tmpdf.columns = ["id_a", "id_b"]
        _interactions_subset = pd.concat([_interactions_subset, tmpdf], axis=1)
        simple_tm0 = pd.DataFrame(
            [rx.split("_") for rx in _interactions_subset_simp.interacting_pair],
            columns=["id_a", "id_b"],
            index=_interactions_subset_simp.index,
        )
        _interactions_subset_simp = pd.concat([_interactions_subset_simp, simple_tm0], axis=1)
        interactions_subset = pd.concat([_interactions_subset_simp, _interactions_subset], axis=0)
    else:
        tm0 = pd.DataFrame(tm0).T
        tm0.columns = ["id_a", "id_b"]
        interactions_subset = pd.concat([interactions_subset, tm0], axis=1)

    # keep only useful genes
    geneid = list(set(list(interactions_subset.id_a) + list(interactions_subset.id_b)))
    if not all([g in adata_subset.var.index for g in geneid]):
        geneid = list(set(list(interactions_subset.gene_a) + list(interactions_subset.gene_b)))
    # create a subet anndata
    adata_subset_tmp = adata_subset[:, adata_subset.var_names.isin(geneid)].copy()
    meta = adata_subset_tmp.obs.copy()
    adata_list, adata_list_alt = {}, {}
    for x in list(set(meta[celltype_key])):
        adata_list[x] = adata_subset_tmp[adata_subset_tmp.obs[celltype_key] == x].copy()
        adata_list_alt[x] = adata_subset[adata_subset.obs[celltype_key] == x].copy()

    # create expression and fraction dataframes.
    adata_list2, adata_list3 = {}, {}
    for x in adata_list:
        adata_list2[x] = celltype_means(adata_list[x], layer)
        adata_list3[x] = celltype_fraction(adata_list[x], layer)
    adata_list2 = pd.DataFrame(adata_list2, index=adata_subset_tmp.var_names)
    adata_list3 = pd.DataFrame(adata_list3, index=adata_subset_tmp.var_names)

    decon_subset = deconvoluted[deconvoluted.complex_name.isin(find_complex(interactions_subset))].copy()

    # if any interactions are actually complexes, extract them from the deconvoluted dataframe.
    if decon_subset.shape[0] > 0:
        decon_subset_expr = decon_subset.groupby("complex_name").apply(lambda r: r[adata_list2.columns].apply(np.mean, axis=0))
        cellfrac = defaultdict(dict)
        zgenes = list(set(decon_subset_expr.index))
        for ct, adat in adata_list_alt.items():
            for zg in zgenes:
                cellfrac[ct][zg] = np.mean(adat[:, adata.var_names.isin(zg.split("_"))].X > 0)
        decon_subset_fraction = pd.DataFrame(cellfrac)
        expr_df = pd.concat([adata_list2, decon_subset_expr])
        fraction_df = pd.concat([adata_list3, decon_subset_fraction])
    else:
        expr_df = adata_list2.copy()
        fraction_df = adata_list3.copy()

    # create edge list
    cells_test = list(set(meta[celltype_key]))
    cell_comb = []
    for c1 in cells_test:
        for c2 in cells_test:
            if remove_self:
                if c1 != c2:
                    cell_comb.append((c1, c2))
            else:
                cell_comb.append((c1, c2))
    cell_comb = list(set(cell_comb))
    cell_type_grid = pd.DataFrame(cell_comb, columns=["source", "target"])
    # create the final dataframe for plotting
    dfx = generate_df(
        interactions_subset=interactions_subset,
        cell_type_grid=cell_type_grid,
        cell_type_means=expr_df,
        cell_type_fractions=fraction_df,
        sep=DEFAULT_SEP,
    )
    # ok form the table for pyCircos
    int_value = dict(zip(lr_interactions.barcode, lr_interactions.y_means))
    int_value = {k: r for k, r in int_value.items() if pd.notnull(r)}
    dfx["interaction_value"] = [int_value[y] if y in int_value else np.nan for y in dfx["barcode"]]
    tmpdf = dfx[["producer", "receiver", "converted_pair", "interaction_value"]].copy()
    tmpdf["interaction_celltype"] = [
        DEFAULT_SEP.join(sorted([a, b, c])) for a, b, c in zip(tmpdf.producer, tmpdf.receiver, tmpdf.converted_pair)
    ]
    celltypes = sorted(list(set(list(tmpdf.producer) + list(tmpdf.receiver))))
    celltype_start_dict = {r: k * gap for k, r in enumerate(celltypes)}
    celltype_end_dict = {r: k + gap for k, r in enumerate(celltypes)}
    interactions = sorted(list(set(tmpdf["interaction_celltype"])))
    interaction_start_dict = {r: k * gap for k, r in enumerate(interactions)}
    interaction_end_dict = {r: k + gap for k, r in enumerate(interactions)}
    tmpdf["from"] = [celltype_start_dict[x] for x in tmpdf.producer]
    tmpdf["to"] = [celltype_end_dict[x] for x in tmpdf.receiver]
    tmpdf["interaction_value"] = [
        j * scale_lw + interaction_start_dict[x] if pd.notnull(j) else np.nan
        for j, x in zip(tmpdf.interaction_value, tmpdf.interaction_celltype)
    ]
    tmpdf["start"] = round(tmpdf["interaction_value"] + tmpdf["from"])
    tmpdf["end"] = round(tmpdf["interaction_value"] + tmpdf["to"])
    if edge_col_dict is None:
        uni_interactions = list(set(tmpdf.converted_pair))
        col_step = 1 / len(uni_interactions)
        start_step = 0
        edge_col_dict = {}
        for i in uni_interactions:
            edge_col_dict[i] = edge_cmap(start_step)
            start_step += col_step
    circle = Gcircle(figsize=figsize)
    if face_col_dict is None:
        if celltype_key + "_colors" in adata.uns:
            if adata.obs[celltype_key].dtype.name == "category":
                face_col_dict = dict(zip(adata.obs[celltype_key].cat.categories, adata.uns[celltype_key + "_colors"]))
            else:
                face_col_dict = dict(zip(list(set(adata.obs[celltype_key])), adata.uns[celltype_key + "_colors"]))
    for i, j in tmpdf.iterrows():
        name = j["producer"]
        col = None if face_col_dict is None else face_col_dict[name]
        arc = Garc(
            arc_id=name,
            size=size,
            interspace=interspace,
            raxis_range=raxis_range,
            labelposition=labelposition,
            label_visible=label_visible,
            facecolor=col,
        )
        circle.add_garc(arc)
    circle.set_garcs(-180, 180)
    for i, j in tmpdf.iterrows():
        if pd.notnull(j["interaction_value"]):
            lr = j["converted_pair"]
            start_size = j["start"] + j["interaction_value"] / scale_lw
            end_size = j["end"] + j["interaction_value"] / scale_lw
            start_size = 1 if start_size < 1 else start_size
            end_size = 1 if end_size < 1 else end_size
            source = (j["producer"], j["start"] - 1, start_size, raxis_range[0] - size)
            destination = (j["receiver"], j["end"] - 1, end_size, raxis_range[0] - size)
            circle.chord_plot(source, destination, edge_col_dict[lr])

    custom_lines = [Line2D([0], [0], color=val, lw=4) for val in edge_col_dict.values()]
    circle.figure.legend(custom_lines, edge_col_dict.keys(), **legend_params)
    return circle



#!/usr/bin/env python
import pkg_resources

try:
    __version__ = pkg_resources.get_distribution("ktplotspy").version
except:  # pragma: no cover
    __version__ = "dev"

DEFAULT_SEP = ">@<"
DEFAULT_SPEC_PAT = "/|:|\\?|\\*|\\+|\\|\\(|\\)|\\/"



#!/usr/bin/env python
from ktplotspy.plot.plot_cpdb import plot_cpdb
from ktplotspy.plot.plot_cpdb_chord import plot_cpdb_chord
from ktplotspy.plot.plot_cpdb_heatmap import plot_cpdb_heatmap

__all__ = ["plot_cpdb", "plot_cpdb_chord", "plot_cpdb_heatmap"]
